# -*- coding: utf-8 -*-

import ast
import datetime
import time
from dateutil.relativedelta import relativedelta
import logging
import json
from passlib.context import CryptContext
import requests
import re
from werkzeug.urls import url_encode
import hmac
from hashlib import sha256
import base64
import uuid

import odoo
from odoo import api, fields, models, tools
from odoo.osv import expression
from odoo.exceptions import AccessDenied, ValidationError, UserError
from odoo.tools import consteq, misc
from odoo.tools.translate import _

from odoo.addons.saas_worker.const import EDU_PREFIX, SAAS_BASE_URL, TIMEOUT, SCHEME, WEBSITE, SAAS_OPS_URL
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.util import (email_on_exception, get_roles_from_ip, ip2coordinates,
                                          humanize_version, get_edu_eol_date)
from odoo.addons.sale_subscription.models.sale_order import SUBSCRIPTION_PROGRESS_STATE

_logger = logging.getLogger(__name__)


HIDDEN_UUIDS = [
    'a0a30d16-6095-11e2-9c70-002590a17fd8',  # apps
    'de1ae970-db51-11e7-be1e-42010a840002',  # iap
    '05a64ced-5b98-488d-a833-a994f9b1dd80',  # test/next
    '8851207e-1ff9-11e0-a147-001cc0f2115e',  # openerp
]

enterprise_crypt_context = CryptContext(
    ['pbkdf2_sha512', 'md5_crypt', 'plaintext'],
    deprecated=['md5_crypt', 'plaintext'],
)

GC_FREEMIUM_WARN = 2      # will warn user after x months
GC_FREEMIUM_CLOSE = 3      # will close contract after x months
assert GC_FREEMIUM_WARN < GC_FREEMIUM_CLOSE


HOSTING_TYPES_SELECTION = [
    ('saas', 'Online'),
    ('paas', 'Odoo.sh'),
    ('premise', 'On Premise')
]

# ----------------------------------------------------------
# OE models
# ----------------------------------------------------------


class OpenerpEnterpriseModuleMapping(models.Model):

    _name = 'openerp.enterprise.module.mapping'
    _description = 'Mapping between Odoo modules and the product to sell them'

    name = fields.Char('Name', required=True)
    product_id = fields.Many2one('product.product', 'Product', required=True, domain=lambda x: [('categ_id', '=', x.env.ref('openerp_enterprise.category_openerp_apps').id)])
    pricelist_ids = fields.Many2many('product.pricelist', 'oe_database_module_mapping_pricelist_rel', 'mapping_id', 'pricelist_id', 'Pricelist', required=True, context={'active_test': False})
    module_id = fields.Many2one('openerp.enterprise.database.module', 'Module', required=True)


class OpenerpEnterpriseModuleVersion(models.Model):
    """ Model that stores the Odoo versions
        The sequences must be ordered by the chronological releases of the versions.
        Meaning, the lower is the sequence, the older is the version.
    """
    _name = 'openerp.enterprise.module.version'
    _description = 'Odoo Release versions'
    _order = 'sequence desc'

    name = fields.Char('Name', required=True, help="Must be a valid version name")
    sequence = fields.Integer('Sequence', required=True, default=0)
    supported = fields.Boolean('Is supported', default=True)
    module_ids = fields.One2many('openerp.enterprise.database.module', 'version_id', 'Applications')
    paid_module_ids = fields.One2many('openerp.enterprise.database.module', 'version_id', 'Paid applications', domain=[('is_sellable', '=', True)])

    @api.model
    def _get_latest_version(self):
        # we force the order to have no issue if someone modifies it
        return self.search([('supported', '=', True)], order='sequence desc', limit=1)

    def apps2products(self, pricelist_id, apps):
        self.ensure_one()
        if not pricelist_id or not apps:
            return self.env['product.product']
        # lower the apps name so the `in` search works
        apps = [a.lower() for a in apps]
        mapping_ids = self.paid_module_ids.filtered(lambda x: x.technical_name in apps).mapped('mapping_ids')
        return mapping_ids.filtered(
                lambda x: pricelist_id in x.pricelist_ids.ids
            ).mapped('product_id')

    def pricelist2modules(self, pricelist_id):
        self.ensure_one()
        if not pricelist_id:
            return self.env['openerp.enterprise.database.module']

        return self.paid_module_ids.filtered(lambda x: pricelist_id in x.mapped('mapping_ids.pricelist_ids').ids)


class OpenerpEnterpriseDatabaseModule(models.Model):

    _name = 'openerp.enterprise.database.module'
    _description = 'Odoo module versionned with its dependencies'
    _order = 'is_extra_integration, sequence, technical_name'

    name = fields.Char('Name', required=True, translate=True)
    technical_name = fields.Char('Technical name', required=True)
    mapping_ids = fields.One2many('openerp.enterprise.module.mapping', 'module_id', 'Mappings')
    dependency_ids = fields.Many2many('openerp.enterprise.database.module', 'oe_database_module_dep_rel', 'module_id', 'dep_id', 'Dependencies')
    version_id = fields.Many2one('openerp.enterprise.module.version', 'Version', required=True)
    is_sellable = fields.Boolean('Is sellable', help="Define if this module is a paid one.", store=False, compute='_compute_is_sellable', search='_search_is_sellable')
    is_extra_integration = fields.Boolean('Extra integration', help="""By checking this checkbox, this module will be displayed
        in the "Extra Integrations" section of the Pricing page""")
    sequence = fields.Integer('Sequence', default=0, help="""This will decide the order in which the app will appear on the pricing page.""")
    success_pack_time = fields.Integer('Success Pack Time', required=True, default=4, help="""Hours required for Success Pack with this module.""")
    website_published = fields.Boolean('Published', default=True, help="Make this module available on the pricing page")

    _sql_constraints = [
        ('module_version_unique', 'UNIQUE(technical_name,version_id)', 'The module technical_name must be unique per version'),
    ]

    def _compute_display_name(self):
        for m in self:
            m.display_name = f'{m.name} ({m.version_id.name})'

    @api.model
    def name_search(self, name, args=[], operator='ilike', limit=100):
        domain = []
        if (name or '').strip():
            if operator in expression.NEGATIVE_TERM_OPERATORS:
                domain = expression.AND([
                    ['&', '&', ('name', operator, name), ('technical_name', operator, name), ('version_id', operator, name)],
                    args,
                ])
            else:
                domain = expression.AND([
                    ['|', '|', ('name', operator, name), ('technical_name', operator, name), ('version_id', operator, name)],
                    args,
                ])
        recs = self.search(domain, limit=limit)
        return [(record.id, record.display_name) for record in recs]

    @api.model_create_multi
    def create(self, vals_list):
        for vals in vals_list:
            # we lower the technical name so for app researches like in the method apps2products we can use the 'in' operator
            vals['technical_name'] = vals['technical_name'].lower().strip()
        return super(OpenerpEnterpriseDatabaseModule, self).create(vals_list)


    def write(self, values):
        if 'technical_name' in values:
            values['technical_name'] = values['technical_name'].lower().strip()

        return super(OpenerpEnterpriseDatabaseModule, self).write(values)

    @api.constrains('dependency_ids')
    def _check_dependency_loop(self):
        return self._check_m2m_recursion('dependency_ids')


    def _compute_is_sellable(self):
        for m in self:
            # if a module has a mapping to a product, then it is considered as a paid module
            m.is_sellable = bool(m.mapping_ids)

    @api.model
    def _search_is_sellable(self, operator, value):
        # search if a module is paid by checking if it has a mapping_id linked
        if (operator, value) in [('=', True), ('!=', False)]:
            return [('mapping_ids', '!=', False)]
        return [('mapping_ids', '=', False)]

    @api.model
    def apps2products(self, pricelist_id, version_name, apps):
        """ This method takes a pricelist, an odoo version and a list of app technical name
            and return the associated products record, not necessarly in order.
        """
        if not pricelist_id or not version_name:
            return self.env['product.product']

        version_id = self.env['openerp.enterprise.module.version'].search([('name', '=', humanize_version(version_name))])
        if not version_id:
            version_id = self.env['openerp.enterprise.module.version']._get_latest_version()
        return version_id.apps2products(pricelist_id, apps)

    @api.model
    def pricelist2apps(self, pricelist_id, version_name):
        """ This method returns the list of technical for modules depending
            on the version and pricelist id we ask for
        """
        if not pricelist_id or not version_name:
            return []
        return self.pricelist2modules(pricelist_id, version_name).mapped('technical_name')

    @api.model
    def pricelist2modules(self, pricelist_id, version_name):
        if not pricelist_id:
            return self.browse()

        Module_version = self.env['openerp.enterprise.module.version']
        version_id = version_name and Module_version.search([('name', '=', humanize_version(version_name))])
        if not version_id:
            version_id = Module_version._get_latest_version()
        return version_id.pricelist2modules(pricelist_id)

    @api.model
    @odoo.tools.ormcache('version_name')
    def pricelist2modules_and_deps(self, version_name):
        Module_version = self.env['openerp.enterprise.module.version']
        version = version_name and Module_version.search([('name', '=', humanize_version(version_name))])
        if not version:
            version = Module_version._get_latest_version()

        mods = version.module_ids

        result = {}
        for mod in mods:
            deps = mod.get_all_dependencies().mapped('technical_name')
            # all modules are an app and so we considere it as dependence of itself
            deps.append(mod.technical_name)
            result[mod.technical_name] = ','.join(list(set(deps)))

        return result

    def module2products(self, pricelist_id):
        if not self or not pricelist_id:
            return self.env['product.product']

        return self.mapped('mapping_ids').filtered(lambda x: pricelist_id in x.pricelist_ids.ids).mapped('product_id')

    def get_all_dependencies(self):
        deps = self.mapped('dependency_ids')
        if deps:
            deps |= deps.get_all_dependencies()
        return deps


class OpenerpEnterpriseDatabaseApp(models.Model):

    _name = "openerp.enterprise.database.app"
    _description = "Odoo app installed on a database"
    _order = 'installed desc, first_date desc'

    database_id = fields.Many2one('openerp.enterprise.database', 'Database', required=True, index=True, ondelete='cascade')
    module_id = fields.Many2one('openerp.enterprise.database.module', 'Application', required=True, index=True)
    name = fields.Char('Technical Name', related='module_id.technical_name', readonly=True)
    shortdesc = fields.Char('Name', related='module_id.name', readonly=True)
    first_date = fields.Date('Installation date')
    last_date = fields.Date('Last known installed date')
    installed = fields.Boolean('Installed')


class OpenerpEnterpriseOauthUser(models.Model):

    _name = "openerp.enterprise.oauth.user"
    _description = "Openerp Enterprise Oauth User"
    _table = 'oe_database_user_rel'
    _rec_name = 'user_login'

    database_id = fields.Many2one('openerp.enterprise.database', 'Database', required=True, ondelete='cascade')
    user_id = fields.Many2one('res.users', 'User', required=True, ondelete='cascade')
    oauth_uid = fields.Integer('OAuth ID', related='user_id.id')
    is_admin = fields.Boolean('Is administrator')
    user_login = fields.Char(related='user_id.login')

    _sql_constraints = [
        ('oauth_user_unique', 'UNIQUE(database_id, user_id)', 'OAuth user already linked to this database.'),
    ]

    @api.autovacuum
    def _gc_oauth_users(self):
        # delete oauth entries for databases that didn't ping in the last 30 days
        self.env.cr.execute("""DELETE FROM oe_database_user_rel o
                                    USING openerp_enterprise_database d
                                WHERE o.database_id = d.id
                                    AND d.last_ping < CURRENT_DATE - integer '30'""")


class OpenerpEnterpriseDatabaseTag(models.Model):

    _name = "openerp.enterprise.database.tag"
    _description = "Customer database tag"

    name = fields.Char('Tag Name', required=True, translate=True)

    _sql_constraints = [
        ('name_uniq', 'unique (name)', "Tag name already exists !"),
    ]


class InvalidTagsError(ValidationError):
    def __init__(self, invalid_tags):
        super().__init__("Invalid tags")
        self.invalid_tags = invalid_tags


class OpenerpEnterpriseDatabase(models.Model):

    _name = "openerp.enterprise.database"
    _description = "Customer database"
    _rec_name = "db_uuid"
    _order = 'last_ping desc'

    def _get_unlink_token(self, expiration=86400):
        self.ensure_one()
        db_secret = self.env['ir.config_parameter'].get_param('database.secret')
        h = hmac.new(db_secret.encode(), digestmod=sha256)
        max_ts = int(time.time()) + expiration
        message = '>%s|%s<' % (max_ts, self.db_uuid)
        h.update(message.encode())
        return '%s.%s' % (max_ts, base64.b64encode(h.digest()).decode())

    def _check_unlink_token(self, token):
        self.ensure_one()
        max_ts, signature = token.split('.')
        db_secret = self.env['ir.config_parameter'].get_param('database.secret')
        h = hmac.new(db_secret.encode(), digestmod=sha256)
        message = '>%s|%s<' % (max_ts, self.db_uuid)
        h.update(message.encode())
        return hmac.compare_digest(base64.b64decode(signature), h.digest()) and int(time.time()) < int(max_ts)

    def _get_unlink_url(self):
        params = {
            'db_uuid': self.db_uuid,
            'token': self._get_unlink_token(),
        }
        return "%s/publisher-warranty/unlink?%s" % (WEBSITE, url_encode(params))


    def is_saas(self):
        # kept for retro-compatibility
        self.ensure_one()
        return self.hosted_on_saas

    def is_activated(self):
        self.ensure_one()
        return not bool(self.confirm_token)

    def _is_edu(self):
        self.ensure_one()
        if self.db_name:
            return self.db_name.startswith(EDU_PREFIX)
        return False

    @api.depends('last_ip')
    def _compute_hosting(self):
        # compute the hosting type using the ip address
        for r in self:
            roles = get_roles_from_ip(r.last_ip)
            if 'paas_worker' in roles:
                r.hosting = 'paas'
            elif 'trial' in roles:
                r.hosting = 'saas'
            else:
                r.hosting = 'premise'

    @api.depends('info')
    def _extract_info(self):
        for db in self:
            try:
                info = json.loads(db.last_ping_id.info or '{}')
            except ValueError:
                info = {}

            if info.get('main_company_country'):
                db.company_country = self.env['res.country'].search([('code', '=', info['main_company_country'])], limit=1)
            else:
                db.company_country = False

            db.hosted_on_saas = db.hosting == 'saas'
            db.backup_size = info.get('backup_size', False)
            db.fs_size = info.get('fs_size', False)
            db.total_size = info.get('total_size', 0)
            db.iot_box_count = info.get('IoTBox', 0)
            maintenance = info.get('maintenance', {})
            modules = maintenance.get('modules', {})
            db.maintenance_lines_count = sum(modules.values())
            db.maintenance_modules = ', '.join(m for m in modules.keys() if modules[m])
            db.maintenance_errors = ', '.join(maintenance.get('errors', []))
            db.has_base_automation = info.get('has_base_automation', False)
            db.is_neutralized = info.get('is_neutralized', False)

    def templatize(self, values):
        self.ensure_one()
        if 'template_logo' in values:
            logo = values.pop('template_logo')
            if logo:
                values['template_logo'] = base64.encodebytes(logo.stream.read())
        values = {
            k: v for k, v in values.items()
            if k.startswith('template_') or k == 'published_for'
        }

        Template = self.env["openerp.enterprise.database.template"]
        if not self.template_ids:
            values['database_id'] = self.id
            Template.create(values)
        else:
            self.template_ids.write(values)

        self.template_ids.flush_recordset(['published_for'])
        return self.template_ids

    db_uuid = fields.Char("Database UUID", required=True, readonly=True)  # UNIQUE INDEX, no need for index=True
    db_name = fields.Char('Database Name', related='last_ping_id.db_name', store=True)
    create_date = fields.Datetime('Created At', readonly=True)
    last_valid_date = fields.Datetime('Validity Date of Last Ping')
    date_valid = fields.Datetime('Validity Date', compute='_compute_date_valid')
    tag_ids = fields.Many2many('openerp.enterprise.database.tag', string="Tags")

    # ping fields
    ping_ids = fields.One2many("openerp.enterprise.ping", "database", "Pings")
    last_ping_id = fields.Many2one('openerp.enterprise.ping', compute='_compute_last_ping',
        store=True, index=True, string="Last Ping")
    first_ping = fields.Date("First Ping", compute='_compute_ping_dates', store=True)
    last_ping = fields.Date("Latest Ping", compute='_compute_ping_dates', store=True)
    age = fields.Integer("Age", aggregator="avg", compute='_compute_ping_dates', store=True)
    alive = fields.Boolean("Alive (last ping 1<month)", compute='_compute_alive')

    # nbr user and quantity fields
    nbr_users = fields.Integer('Total Users', related='last_ping_id.nbr_users', readonly=True, store=True, help='Total number of active users, incl. portal users')
    nbr_active_users = fields.Integer('Active Users', related='last_ping_id.nbr_active_users', readonly=True, store=True, help='Users that signed in during the past 15 days')
    nbr_share_users = fields.Integer('Portal Users', related='last_ping_id.nbr_share_users', readonly=True, store=True)
    nbr_active_share_users = fields.Integer('Active Portal Users', related='last_ping_id.nbr_active_share_users', readonly=True, store=True, help='Portal users that signed in during the past 15 days')
    nbr_active_regular_users = fields.Integer('Active Regular Users', compute='_compute_quantities', store=True, help='Regular users that signed in during the past 15 days')
    quantity = fields.Integer('Regular Users', compute='_compute_quantities', store=True)

    # global infos
    language = fields.Char('Language', related='last_ping_id.language', readonly=True, store=True)
    version = fields.Char('Version', related='last_ping_id.version', readonly=True, store=True)
    last_ip = fields.Char('Last IP', related='last_ping_id.ip', readonly=True, store=True)
    country = fields.Many2one(related='last_ping_id.country', string='Hosting Location', readonly=True, store=True)
    city = fields.Char('City', related='last_ping_id.city', readonly=True, store=True)
    info = fields.Text('Info', related='last_ping_id.info', readonly=True, store=True)

    # Restricted Company Info, only available for debugging (tech support), normal users
    # shouldn't access this info, which could include personal data that cannot be used for
    # marketing purposes (on-premise community editions, for example)
    company_name = fields.Char('Company Name', related='last_ping_id.company_name', readonly=True, store=True,
        groups="openerp_enterprise.group_support_technical")
    company_email = fields.Char('Company Email', related='last_ping_id.company_email', readonly=True, store=True,
        groups="openerp_enterprise.group_support_technical")
    company_phone = fields.Char('Company Phone', related='last_ping_id.company_phone', readonly=True, store=True,
        groups="openerp_enterprise.group_support_technical")

    hosting = fields.Selection(HOSTING_TYPES_SELECTION, 'Hosted on', store=True, compute='_compute_hosting')
    hosted_on_saas = fields.Boolean('Hosted On SaaS', compute='_extract_info')
    backup_size = fields.Char('Backup Size', compute='_extract_info', help='Only available if hosted on SaaS')
    fs_size = fields.Char('Filestore Size', compute='_extract_info', help='Only available if hosted on SaaS')
    total_size = fields.Integer('Total Size', compute='_extract_info', help='Only available if hosted on SaaS')
    parent_id = fields.Many2one('openerp.enterprise.database', 'Parent Database', help='Source database if this database is a duplicate')
    child_ids = fields.One2many('openerp.enterprise.database', 'parent_id', string='Child databases')
    company_country = fields.Many2one('res.country', compute='_extract_info', string="Company country")
    iot_box_count = fields.Integer('IoT box count', compute='_extract_info')
    maintenance_lines_count = fields.Integer("Custom Line of code", compute='_extract_info')
    maintenance_modules = fields.Char("Maintenance modules", compute='_extract_info')
    maintenance_errors = fields.Char("Maintenance counting errors", compute='_extract_info')
    has_base_automation = fields.Boolean(compute="_extract_info")
    is_neutralized = fields.Boolean(compute="_extract_info")

    url = fields.Char('URL', compute='_get_url', inverse='_set_url', store=True, help='If left empty, will use the url of last ping')
    _url = fields.Char()

    website_homepage_modified = fields.Boolean('Homepage Modified', related='last_ping_id.website_homepage_modified', readonly=True, store=True)
    last_cnx_date = fields.Datetime('Last Connection Date', related='last_ping_id.last_cnx_date', readonly=True, store=True)
    expire_date = fields.Datetime('Expire Date', related='last_ping_id.expire_date', readonly=True, store=True)
    extra_apps = fields.Boolean('Extra Apps', related='last_ping_id.extra_apps', readonly=True, store=True)
    spoof_ids = fields.One2many('openerp.enterprise.spoof', 'db_id', 'Support Connections', readonly=True)

    # Template and subscription
    subscription_id = fields.Many2one('sale.order', 'Contract', index=True, domain=[('is_subscription', '=', True)])
    subscription_code = fields.Char(related='subscription_id.client_order_ref', string='Contract Code')
    registered_user_ids = fields.One2many('openerp.enterprise.oauth.user', 'database_id', 'OAuth Users')
    registered_user_admin_ids = fields.One2many(
        'openerp.enterprise.oauth.user', 'database_id',
        string='Admin OAuth Users', domain=[('is_admin', '=', True)]
    )

    location = fields.Selection([
        ('online', 'Online'),
        ('onsite', 'On Site'),
    ], 'Location')
    mode = fields.Char('Mode', help="either: saas, trial, trial15days, onsite, paid")

    ab_message = fields.Char('Announcement Bar Message')
    app_ids = fields.One2many('openerp.enterprise.database.app', 'database_id', 'Installed Apps')
    confirm_token = fields.Char('Confirm Token')
    database_user_ids = fields.One2many('openerp.enterprise.database.user', 'database_id', 'Databases users')

    visibility = fields.Selection([('customer', 'Customer'),
        ('hidden', 'Hidden'),
        ('support', 'Support')], string='Visibility', default='customer')

    allow_free_domain = fields.Boolean('Allow Free Domain Name', help='After careful review, allow this database to register a free domain name, ignoring usual checks.')
    template_ids = fields.One2many('openerp.enterprise.database.template', 'database_id')

    _sql_constraints = [
        ('db_uuid_unique', 'unique(db_uuid)', 'DB UUID must be unique (by definition)'),
    ]

    def action_connect_support_page(self):
        self.ensure_one()
        if self.hosting not in ("saas", "paas"):
            return {}  # no support page for those dbs

        support_page_url = f"{self.url}/_odoo/support"

        if self.env.context.get("task_id"):
            task = self.env["project.task"].search([('id', '=', self.env.context["task_id"])])
            if task:
                # for saas/paas, if we have the support ticket (task) as context
                # then auto-login with a pre-defined reason
                support_page_url = f"{self.url}/_odoo/support/login-with-odoo?reason=opw-{task.id}"

        return {
            'type': 'ir.actions.act_url',
            'target': 'new',
            'url': support_page_url,
        }

    @api.constrains('parent_id')
    def _check_parent_recursion(self):
        if not self._check_recursion():
            raise ValidationError(_("You cannot set a database as its own parent."))

    def _compute_alive(self):
        last_month = datetime.date.today() - datetime.timedelta(days=30)
        for r in self:
            r.alive = r.last_ping and fields.Date.from_string(r.last_ping) > last_month


    @api.depends('ping_ids', 'ping_ids.ping_time')
    def _compute_last_ping(self):
        for database in self:
            database.last_ping_id = False
            if database.id:
                ping_ids = self.last_ping_id.search([('database', '=', database.id)], limit=1, order="ping_time desc, id desc")
            else:
                # NewId computation for on_change
                ping_ids = sorted(self.ping_ids, key=lambda x: x.ping_time, reverse=True)

            if ping_ids:
                database.last_ping_id = ping_ids[0]

    @api.depends('last_ping_id')
    def _compute_ping_dates(self):
        self.env['openerp.enterprise.ping'].flush_model(['database', 'ping_time'])
        self.env.cr.execute("""
            SELECT database, min(ping_time), max(ping_time)
            FROM openerp_enterprise_ping
            WHERE database IN %s
            GROUP BY database""", (tuple(self.ids),))

        result = dict.fromkeys(self.ids, dict())
        for database_id, min_ping_time, max_ping_time in self.env.cr.fetchall():
            result[database_id] = {
                'first_ping': min_ping_time.date(),
                'last_ping': max_ping_time.date(),
                'age': (max_ping_time - min_ping_time).days
            }

        for database in self:
            database.update(result.get(database.id) or {'first_ping': False, 'last_ping': False, 'age': False})

    @api.depends('last_valid_date')
    def _compute_date_valid(self):
        for database in self:
            if database.subscription_id and database.subscription_id.subscription_state in SUBSCRIPTION_PROGRESS_STATE:
                database.date_valid = database._get_valid_date().get(database.id, False)
            else:
                database.date_valid = database.last_valid_date

    @api.depends('nbr_users', 'nbr_share_users', 'nbr_active_users', 'nbr_active_share_users',
                 'registered_user_ids.user_id')
    def _compute_quantities(self):
        month_ago = fields.Date.to_string(datetime.date.today() + relativedelta(months=-1))
        for database in self:
            pings = self.env['openerp.enterprise.ping'].search([('database', '=', database.id), ('ping_time', '>', month_ago)])
            if not pings:
                last_ping = database.last_ping_id
                if last_ping:
                    pings = last_ping

            user_counts = [int(ping.nbr_users) - int(ping.nbr_share_users) for ping in pings]
            effective_users_count = user_counts[0] if user_counts else 1

            # subtract accounting firm users, those should not be counted as billable users
            # (with a minimal value of 1 user if all users are from accounting firm)
            def _is_accounting_firm(db_user):
                comm_partner = db_user.user_id.commercial_partner_id
                return comm_partner.portal_accounting_firm_level and comm_partner.is_published

            accounting_firm_users_count = len(database.registered_user_ids.filtered(_is_accounting_firm))
            if accounting_firm_users_count:
                effective_users_count = max(effective_users_count - accounting_firm_users_count, 1)

            database.quantity = effective_users_count
            database.nbr_active_regular_users = max(0, pings[0].nbr_active_users - pings[0].nbr_active_share_users) if pings else 0

    @api.depends('last_ping_id', '_url')
    def _get_url(self):
        for database in self:
            database.url = database._url or database.last_ping_id.web_base_url


    def _set_url(self):
        self = self[0]
        _url = self.url or None
        url = self.last_ping_id.web_base_url
        if _url == url:
            _url = None
        self.env.cr.execute("UPDATE openerp_enterprise_database SET _url=%s, url=%s WHERE id=%s",
                            [_url, _url or url, self.id])
        self.invalidate_recordset(['url', '_url'])

    # ORM overrides

    def _compute_display_name(self):
        for database in self:
            name_elements = []
            if database.url:
                name_elements.append(database.url.replace('https://', '').replace('http://', ''))
            if database.db_name:
                name_elements.append(database.db_name)
            name = "#".join(name_elements)
            if not name and database.db_uuid:
                uuid_parts = database.db_uuid.split('-')
                for i in range(len(uuid_parts) - 1):
                    uuid_parts[i] = 'xxxx'
                name = "-".join(uuid_parts)
            database.display_name = name

    def write(self, vals):
        recurring_subscriptions = self.env['sale.order']

        if 'subscription_id' in vals:
            new_subscription = self.env['sale.order'].browse(vals['subscription_id'])
            for database in self:
                reason = f"sub: {new_subscription.display_name}"
                if database.subscription_id and not database._check_spoof_allowed('subscription'):
                    database._check_and_log_spoof('subscription', reason=reason)
                    raise AccessDenied(_("This database is restricted! You need to ask for a staff access authorization to the customer."))
                if database.subscription_id and vals['subscription_id']:
                    new_subscription.message_post(body=_('Database %s(%s) added to this contract.') % (database.db_name or _('N/A'), database.db_uuid or _('N/A')))
                    database.subscription_id.message_post(body=_('Database %s(%s) moved to contract %s.') % (database.db_name or _('N/A'), database.db_uuid or _('N/A'), new_subscription.name or _('N/A')),)
                    recurring_subscriptions |= database.subscription_id | new_subscription
                elif database.subscription_id:
                    database.subscription_id.message_post(body=_('Database %s(%s) removed from this contract.') % (database.db_name or _('N/A'), database.db_uuid or _('N/A')))
                    recurring_subscriptions |= database.subscription_id
                elif vals['subscription_id']:
                    new_subscription.message_post(body=_('Database %s(%s) added to this contract.') % (database.db_name or _('N/A'), database.db_uuid or _('N/A')))
                    recurring_subscriptions |= new_subscription
        result = super(OpenerpEnterpriseDatabase, self).write(vals)

        if 'subscription_id' in vals:
            # validate access and log subscription change against the new subscription
            for database in self:
                reason = f"sub: {database.subscription_id.display_name}"
                if not database._check_and_log_spoof('subscription', reason=reason):
                    raise AccessDenied(_("This database is restricted! You need to ask for a staff access authorization to the customer."))

        if {'hosting', 'subscription_id'} & vals.keys():
            self._update_subscription_hosting()

        if recurring_subscriptions:
            # Add maintenance_subscription since they may need changes when database is linked or unlinked
            recurring_subscriptions |= recurring_subscriptions.mapped('maintenance_subscription_id')
            # Only set recurring invoice lines for contracts with interval period less than a month. Other contracts must pay offplans, and offplans are based on the recurring invoice lines.
            recurring_subscriptions.filtered(lambda sub: sub.plan_id.billing_period_unit == 'month').set_apps_recurring_lines()
        return result

    # Business and Helpers

    @misc.cache()
    def _get_self_dbuuid(self):
        return self.env['ir.config_parameter'].sudo().get_param('database.uuid')

    @api.model
    def get_instances(self):
        # Keep retro-compatibility for awesome timesheets
        return self._get_instances()

    @api.model
    def _get_instances(self, db_ids=None, alive=False):
        """
        List all the databases the current user can access.

        :param list(integer) db_ids: fore the inclusion of these databases in the result

        :return: a list of all the databases on which the current user is a registered
                 user (through OAuth). The list contains one dict per database with some
                 generic info and the following entries:
                     - is_admin: True if the current user is that database's oauth admin
                     - is_customer: True if the current user's partner (or their company) is set in
                                    one of these fields: partner_id, enterprise_final_customer_id
                     - subscription info (has_subscription, subscription_state, subscription_total)
                     - is_trial: True if the database has no subscription and is not a one app free
                     - is_edu: True if the database has the edu- prefix
        :rtype: list(dict)
        """
        DatabaseSudo = self.env['openerp.enterprise.database'].sudo()
        DbUserSudo = DatabaseSudo.registered_user_ids

        result = []
        user = self.env.user
        db_ids = list(db_ids) if db_ids else []
        databases = DatabaseSudo.browse(db_ids)
        db_domain = [('registered_user_ids.user_id', 'in', [self._uid])]
        if alive:
            last_month = datetime.date.today() - datetime.timedelta(days=30)
            db_domain.append(('last_ping', '>', fields.Date.to_string(last_month)))
        if self._uid:
            # ignore internal databases
            db_domain.append(('db_uuid', 'not in', HIDDEN_UUIDS))
            databases |= self.sudo().search(db_domain, order='db_name ASC')

        dbuser_per_db = {x.database_id.id: x for x in DbUserSudo.search([
            ('user_id', '=', user.id),
            ('database_id', 'in', databases.ids),
        ])}
        for database in databases:
            partner = database.subscription_id.partner_id
            dbuser = dbuser_per_db.get(database.id, False)

            # user properties and roles
            is_saas = database.hosted_on_saas
            is_admin = dbuser.is_admin if dbuser else False
            is_trial = not bool(database.subscription_id) and database.extra_apps

            final_cust = database.subscription_id.enterprise_final_customer_id
            customers = (partner.commercial_partner_id, final_cust.commercial_partner_id)
            is_customer = user.partner_id.commercial_partner_id in customers

            is_registered_user = dbuser is not False

            # instance values
            values = {
                'db_id': database.id,
                'db_name': database.db_name or database.company_name or database.url.lower(),
                'version': database.version,
                'company_name': database.company_name or database.db_name or database.url.lower(),
                'uuid': database.db_uuid,
                'parent_uuid': database.parent_id.db_uuid,
                'url': database.url.lower() if database.url else False,
                'tags': database.tag_ids.mapped("name"),
                'is_edu': database.db_name and database.db_name.startswith(EDU_PREFIX),
                'is_admin': is_admin,
                'is_customer': is_customer,
                'is_saas': is_saas,
                'has_sub': bool(database.subscription_id),
                'sub_total': database.subscription_id.recurring_monthly if database.subscription_id else False,
                'total_size': database.total_size,
                'is_registered_user': is_registered_user,
                'is_trial': is_trial,
                'is_paid': bool(database.subscription_id.amount_total),
                'auth_url': "%s/web" % (database.url.lower(),) if database.url else False,
                'visibility': database.visibility,
                'allow_free_domain': database.allow_free_domain,
                'is_activated': database.is_activated(),
                'is_neutralized': database.is_neutralized,
            }
            if values['is_edu']:
                values.update(edu_expire_date=database.expire_date or '',  # from last ping
                              edu_eol_date=get_edu_eol_date(values['version']) or '')

            result.append(values)
        return result

    @api.model
    def get_databases(self, alive=False):
        """
        List the paying/supported databases the current user has access to.

        Filter the result of _get_instances to only include paying databases. A database is
        considered as paying if it has a non-free subscription, supported databases include
        edu- databases. See the _get_instances docstring for more information.

        :param boolean alive: only list databases that pinged in the last 30 days

        :return: a list of all the supported databases on which the current user is a registered
                 user (through OAuth).
        :rtype: list(dict)
        """
        instances = self._get_instances(alive=alive)
        return [x for x in instances if ((x['is_customer'] and x['sub_total'] > 0) or x['is_edu']) and x['is_saas']]


    def _unlink_subscription(self):
        self.write({'subscription_id': False})

    def disable(self):
        self.ensure_one()
        reason = f"db: {self.db_name}"
        if not self._check_and_log_spoof('drop', reason=reason):
            raise AccessDenied(_("This database is restricted! You need to ask for a staff access authorization to the customer."))
        self.registered_user_ids.unlink()
        return self.write({'subscription_id': False})

    def unsubscribe(self, delete_shortcut):
        self.ensure_one()
        data = self.subscription_id.sudo().read(['recurring_monthly', 'partner_id'])[0]
        user_partner_id = self.env.user.partner_id
        if data['partner_id'] in (user_partner_id.id, user_partner_id.commercial_partner_id.id):
            raise AccessDenied()
        if data['recurring_monthly'] == 0:
            self.sudo().write({'subscription_id': False})
            if delete_shortcut:
                self.sudo().registered_user_ids.unlink()
        return True

    @api.model
    @email_on_exception()
    def _cron_gc_freemiums(self):
        template_warn_gc = self.env.ref('openerp_enterprise.gc_warn_freemium_email')
        warn = fields.Datetime.to_string(datetime.datetime.now() - relativedelta(months=GC_FREEMIUM_WARN))
        close = fields.Datetime.to_string(datetime.datetime.now() - relativedelta(months=GC_FREEMIUM_CLOSE))

        base_domain = [
            ('subscription_id.subscription_state', '=', '3_progress'),
            ('subscription_id.recurring_monthly', '=', 0),

            # You may thing that we can use `enterprise_saas` field, but that's not true.
            # This fiekd mean that *any* database is on saas, while we need contracts
            # with *all* databases on saas.
            # ('analytic_account.enterprise_saas', '=', True)
        ]

        # close contracts
        domain = base_domain + [('last_cnx_date', '<', close)]
        saas_databases = self.search(domain).filtered(lambda db: db.is_saas())

        subscription_to_close = self.env['sale.order']
        for database in saas_databases:
            subscription = database.subscription_id
            if all(d in saas_databases for d in subscription.database_ids):
                subscription_to_close |= subscription

        for subscription in subscription_to_close:
            subscription.set_close()
            self.env.cr.commit()

        # warn users
        domain = base_domain + [('last_cnx_date', '<', warn), ('last_cnx_date', '>', close)]
        for db in self.search(domain):
            if not db.is_saas():
                continue

            email_context = dict(self._context, **{
                'GC_FREEMIUM_WARN': GC_FREEMIUM_WARN,
                'db': db,
            })
            template_warn_gc.sudo().with_context(email_context).send_mail(db.subscription_id.id)
            self.env.cr.commit()

    @api.model
    def _cron_reminder_expiration_one_app_free(self):
        template_reminder = self.env.ref('openerp_enterprise.reminder_expiration_one_app_free_email').sudo()
        warn = datetime.date.today() + relativedelta(days=14)
        two_weeks_ago = datetime.date.today() - relativedelta(days=15)
        two_months_ago = datetime.date.today() - relativedelta(months=2)
        for database in self.search([('subscription_id', '=', False),
                                     ('extra_apps', '=', False),
                                     ('hosting', '=', 'saas'),
                                     ('create_date', '<', two_months_ago), # don't warn fresh trials!
                                     ('last_ping', '>=', two_weeks_ago),
                                     ('expire_date', '>=', warn), ('expire_date', '<', warn + relativedelta(days=1))]):
            email_ctx = dict(self._context, **{
                'db': database,
                'days_left': 14,
                'email_to': database.company_email,
            })
            try:
                template_reminder.with_context(email_ctx).send_mail(database.subscription_id.id)
                _logger.info("1-app free expiration reminder sent to %s for %s", database.company_email, database)
            except Exception:
                _logger.exception("Failed to send 1-app free expiration reminder for %s", database)

    @api.model
    def _cron_reminder_trial_flow(self):
        # Warn between 24-48 before expiration
        domain = [
            ('extra_apps', '=', True),
            ('hosting', '=', 'saas'),
            '|',
            ('subscription_id', '=', False),
            ('subscription_id.subscription_state', 'in', ['1_draft', '2_renewal']),
        ]

        template_pre_expiration = self.env.ref('openerp_enterprise.reminder_trial_pre_expiration')
        tomorrow_date_str = fields.Date.to_string(datetime.datetime.now() + relativedelta(days=1))
        in_two_days_date_str = fields.Date.to_string(datetime.datetime.now() + relativedelta(days=2))

        for database in self.search(expression.AND([domain, [
            ('expire_date', '>=', tomorrow_date_str),
            ('expire_date', '<', in_two_days_date_str),
        ]])):
            upgrade_link = WEBSITE + '/odoo-enterprise/upgrade-online/%s' % database.db_uuid
            admin_emails = database.registered_user_ids.filtered('is_admin').mapped('user_id.email')
            email_context = dict(self._context, **{
                'email_to': ','.join(admin_emails),
                'upgrade_link': upgrade_link,
            })
            template_pre_expiration.sudo().with_context(email_context).send_mail(database.id)

        # Warn in the 24h following the expiration
        template_post_expiration = self.env.ref('openerp_enterprise.reminder_trial_post_expiration')
        yesterday_date_str = fields.Date.to_string(datetime.datetime.now() - relativedelta(days=1))
        today_date_str = fields.Date.to_string(datetime.datetime.now())

        for database in self.search(expression.AND([domain, [
            ('expire_date', '>=', yesterday_date_str),
            ('expire_date', '<', today_date_str),
        ]])):
            admin_emails = database.registered_user_ids.filtered('is_admin').mapped('user_id.email')
            email_context = dict(self._context, **{
                'email_to': ','.join(admin_emails),
            })
            template_post_expiration.sudo().with_context(email_context).send_mail(database.id)

    @api.model
    def _cron_garbage_collect_saas_draft_contract(self, batch_size=1000, limit=None):
        """ Delete contracts in draft state and older than 1 day

        :param batch_size: number of subscriptions to delete at once (before commiting)
        :param limit: if specific, the maximum number of subscriptions to delete
        """
        saas_draft_query = """
            SELECT sub.id
              FROM sale_order sub
             WHERE sub.is_subscription = true
               AND sub.origin ILIKE %s
               AND sub.state in ('draft', 'cancel')
               AND sub.subscription_state = '1_draft'
               AND sub.write_date < NOW() - interval '1 day'
               AND NOT EXISTS(SELECT NULL from account_move_line l where l.subscription_id = sub.id)
               AND NOT EXISTS(SELECT NULL FROM sale_order_transaction_rel sotx WHERE sotx.sale_order_id = sub.id)
        """
        saas_draft_query_args = ("SaaS Subscription %",)
        if limit:
            # when applying a limit, get the oldest draft subscriptions first
            saas_draft_query += "ORDER BY sub.id ASC LIMIT %s"
            saas_draft_query_args += (limit,)
        self.env.cr.execute(saas_draft_query, saas_draft_query_args)
        # use unlink to remove related mail_activity, mail_followers, mail_message, ...
        saas_draft_sub_ids = sum(self.env.cr.fetchall(), ())
        for sub_ids in self.env.cr.split_for_in_conditions(saas_draft_sub_ids, size=batch_size):
            so = self.env['sale.order'].sudo().browse(sub_ids)
            self.env['openerp.enterprise.spoof'].search([('subscription_id', 'in', so.ids), ('action', '=', 'subscription')]).unlink()
            self.env['openerp.enterprise.spoof'].search([('subscription_id', 'in', so.ids), ('action', '!=', 'subscription')]).write({'subscription_id': False})
            so.order_line.unlink()
            so.unlink()
            self.env.cr.commit()

    def _update_subscription_hosting(self):
        for database in self:
            subscription = database.subscription_id
            if not subscription:
                continue

            if database.hosting != subscription.hosting:
                if database.hosting == 'saas' and subscription.hosting == 'premise':
                    subscription.hosting = 'saas'
                elif database.hosting == 'paas' and subscription.hosting in ('premise', 'saas'):
                    subscription.hosting = 'paas'

    def _get_valid_date(self, code=False):
        """ Enterprise contracts: validity date is the one from the contract
            Partnership contract: increase by 30 days everytime it's requested by code
            Monthly contracts: add 30 days so we don't trigger expiration message on Odoo.sh
            No contract (Enterprise): 30 days after first enterprise ping
            No contract (Community): no expiration date
        """
        result = {}
        for database in self:
            date = False
            subscription = database.subscription_id

            if subscription and subscription.enterprise_is_enterprise:
                date = subscription.next_invoice_date
                if subscription.next_invoice_date == subscription.start_date:
                    date += relativedelta(days=15)
                elif database.hosting in ('paas', 'premise') and subscription.plan_id.billing_period_unit == 'year':
                    date += relativedelta(days=15)

                # check if upsell limit
                if date and subscription.upsell_date:
                    date = min(date, subscription.upsell_date)

                # check if Suspsend invoicing
                if date and subscription.subscription_state == '4_paused':
                    date = date + relativedelta(days=30)

            elif subscription and subscription.enterprise_is_partnership:
                if code and subscription.client_order_ref == code:
                    date = datetime.datetime.now() + relativedelta(days=30)

            # Since 13.2, expiration_panel.js will try to fetch the
            # ir.config_parameter 'database.expiration_date', resulting in
            # warnings wrongly triggered at multiple places if it's set improperly.
            # This is problematic with 1-app free databases and that's why we force
            # the date_valid if the DB has no subscription and is hosted on our SaaS.
            elif database.hosting == 'saas':
                date = database.expire_date
            # If a subscription not valid is set, do not change the valid date
            elif not subscription:
                first_ent_ping = self.env['openerp.enterprise.ping'].search([('database', '=', database.id), ('version', 'ilike', 'e')], limit=1, order="id asc")
                date = first_ent_ping and first_ent_ping.ping_time + relativedelta(days=30)
            # if the subscription template is a monthly and the contract has a payment token id
            if subscription and date and subscription.plan_id.billing_period_unit == 'month' and subscription.payment_token_id:
                if subscription.subscription_state in ('3_progress', '4_paused') and not subscription.to_renew:  # if the subscription is running
                    # we add 38 days (a month + 7 days of latency for the ping)
                    date += relativedelta(days=38)
                elif subscription.subscription_state in ('3_progress', '4_paused') and subscription.to_renew and 'paas' in subscription.database_ids.mapped('hosting'):
                    # we allow our databases hosted in paas to access their databases 15 extra days
                    # if a payment token is linked to the subscription and that the subscription is running and needs to be renewed
                    date += relativedelta(days=15)

            mnt_subscription = subscription.maintenance_subscription_id
            if (date and mnt_subscription.next_invoice_date
                    and mnt_subscription.subscription_state in ('3_progress', '4_paused')):
                if (mnt_subscription.plan_id.billing_period_unit == 'month'
                        and mnt_subscription.recurring_monthly > 0.0
                        and (not mnt_subscription.payment_token_id or mnt_subscription.to_renew)):
                    # If the maintenance subscription has a valid token, use the main subscription date
                    # otherwise gives 15 days to pay and add a token on the maintenance subscription
                    date = min(date, mnt_subscription.next_invoice_date + relativedelta(days=15))
                if mnt_subscription.plan_id.billing_period_unit == 'year' and mnt_subscription.to_renew:
                    # If the maintenance subscription need to be renewed use the min(upsell_date and recurring_next_date)
                    date = min(date, mnt_subscription.next_invoice_date + relativedelta(days=15))
                    if mnt_subscription.upsell_date:
                        date = min(date, mnt_subscription.upsell_date)

            new_date = date and (date + datetime.timedelta(days=1))
            result[database.id] = new_date
        return result

    def update_valid_date(self, code=False):
        valid_dates = self._get_valid_date(code)
        for database in self:
            new_date = valid_dates.get(database.id, False)
            if not database.last_valid_date or (new_date and new_date != database.last_valid_date.date()):
                database.last_valid_date = new_date

    def _is_user_admin(self, user):
        """ Return if a `res.user` record is an oauth admin of the `self` database """
        self.ensure_one()
        return bool(self.registered_user_ids.filtered(lambda x: x.user_id == user and x.is_admin))

    def set_tags(self, tags):
        tags = [tag.lower() for tag in tags]
        bad_tags = [tag for tag in tags if not tag or len(tag) > 30]
        if bad_tags:
            raise InvalidTagsError(bad_tags)

        existing_tags = self.tag_ids.search([('name', 'in', tags)])
        new_tags = set(tags) - set(existing_tags.mapped('name'))

        self.write({
            'tag_ids':
                [(6, 0, existing_tags.ids)] +
                [(0, 0, {'name': tag}) for tag in new_tags],
        })

    def update_tags(self, tag_changes):
        Tag = self.env['openerp.enterprise.database.tag']

        tags = [tc['name'] for tc in tag_changes]
        bad_tags = [tag for tag in tags if not tag or len(tag) > 30]
        if bad_tags:
            raise InvalidTagsError(bad_tags)

        existing_tags = {t.name: t.id for t in Tag.search([('name', 'in', tags)])}

        tag_commands = []
        for tc in tag_changes:
            name = tc['name']
            tag_id = existing_tags.get(name)
            # add/link a tag
            if tc['action'] == 'add':
                if tag_id:
                    tag_commands.append(fields.Command.link(tag_id))
                else:
                    tag_commands.append(fields.Command.create(dict(name=name)))
            # unlink a tag
            elif tc['action'] == 'del':
                tag_commands.append(fields.Command.unlink(tag_id))

        self.write({
            'tag_ids': tag_commands,
        })


class OpenerpEnterpriseDatabaseTemplate(models.Model):
    _name = "openerp.enterprise.database.template"
    _inherits = {'openerp.enterprise.database': 'database_id'}
    _description = "Shared customer database"
    _rec_name = 'db_name'

    database_id = fields.Many2one('openerp.enterprise.database', required=True, ondelete="cascade")
    published_for = fields.Selection(
        [
            ('student', 'For Student'),
            ('teacher', 'For Teacher'),
            ('all', 'For Students & Teachers'),
        ],
        'Database Template Published for')
    template_name = fields.Char('Template Name')
    template_description = fields.Html('Template Description')
    template_author = fields.Char('Template Author')
    template_company = fields.Char('Template School')
    template_logo = fields.Image('Template Logo', max_width=512, max_height=512)
    template_access_token = fields.Char('Security Token', copy=False)
    # Allow to show binary content (template_logo) from published template
    website_published = fields.Selection(related='published_for', string="Website Published")

    _sql_constraints = [
        ('database_id', 'unique (database_id)', "database_id should be unique"),
    ]

    @api.model
    def get_templates(self, domain=None, search=''):
        a_week_ago = fields.Datetime.to_string(datetime.datetime.now() - relativedelta(days=8))
        base_domain = [
            ('published_for', '!=', False),
            ('database_id.last_ping', '>', a_week_ago),
            ('database_id.hosting', '=', 'saas')
        ]
        if search:
            base_domain += [
                '|', ('template_author', 'ilike', search),
                '|', ('template_name', 'ilike', search),
                '|', ('template_company', 'ilike', search),
                     ('template_description', 'ilike', search)
            ]
        if not domain:
            domain = base_domain
        else:
            domain = expression.AND([base_domain, domain])

        templates = self.search(domain)
        return templates

    def _verify_template_token(self, access_token):
        self.ensure_one()
        if not self.template_access_token:
            return False
        return consteq(self.template_access_token, access_token)

    def _portal_ensure_token(self):
        """ Get the current record access token """
        self.ensure_one()
        if not self.template_access_token:
            self.sudo().write({'template_access_token': str(uuid.uuid4())})
        return self.template_access_token

    def _get_access_url(self):
        self.ensure_one()
        params = url_encode({'token': self._portal_ensure_token()})
        base_url = self.get_base_url()
        return f"{base_url}/odoo-enterprise/template/{self.id}?{params}"

class OpenerpEnterprisePing(models.Model):

    _name = "openerp.enterprise.ping"
    _description = "Customer database telemetry"
    _order = 'ping_time desc'

    database = fields.Many2one('openerp.enterprise.database', string="Database", required=True, index=True, ondelete='cascade')
    # computed inernally
    ping_time = fields.Datetime("Ping time", required=True, index=False, default=fields.Datetime.now)
    ip = fields.Char("IP", required=True)
    # required fields introduced in v6.0
    db_name = fields.Char("Database Name", required=True)
    version = fields.Char("Version", required=True)
    nbr_users = fields.Integer("Total Users", required=True)
    language = fields.Char("Language", required=True)
    # additional fields introduced in v6.1
    nbr_active_users = fields.Integer("Active Users")
    nbr_share_users = fields.Integer("Portal Users")
    nbr_active_share_users = fields.Integer("Active Portal Users")
    nbr_regular_users = fields.Integer("Regular Users", compute="_compute_users")
    nbr_active_regular_users = fields.Integer("Active Regular Users", compute="_compute_users")
    # additional fields introduced in v6.1 revno 4187
    # since version 7.0 they are sent as company_<field> instead of <field>
    company_name = fields.Char('Company Name')
    company_phone = fields.Char('Company Phone')
    company_email = fields.Char('Company Email')
    # info field all the values not in colums as json introduced in v6.1 revno 4187
    info = fields.Text("Extra infos")
    # base_url introduced in 7.0
    web_base_url = fields.Char('URL')

    website_homepage_modified = fields.Boolean('Homepage modified')
    last_cnx_date = fields.Datetime('Last Connection Date')
    expire_date = fields.Datetime('Expire Date')
    extra_apps = fields.Boolean('Extra Apps')
    # computed field based on the ip
    country = fields.Many2one('res.country', compute='_compute_country', string="Hosting Location", readonly=True, store=True)
    city = fields.Char('City', compute='_compute_city', readonly=True)

    @api.depends('nbr_users', 'nbr_active_users', 'nbr_share_users', 'nbr_active_share_users')
    def _compute_users(self):
        for ping in self:
            ping.nbr_regular_users = ping.nbr_users - ping.nbr_share_users
            ping.nbr_active_regular_users = ping.nbr_active_users - ping.nbr_active_share_users


    @api.depends('ip')
    def _compute_country(self):
        for ping in self:
            coordinates = ip2coordinates(ping.ip)
            if not coordinates:
                continue

            country_code = coordinates['country_code']
            country = self.env['res.country'].search([('code', '=', country_code)], limit=1)
            ping.country = country

    @api.depends('ip')
    def _compute_city(self):
        for ping in self:
            coordinates = ip2coordinates(ping.ip)
            if not coordinates:
                continue
            ping.city = coordinates['city']

    def _compute_display_name(self):
        for ping in self:
            ping.display_name = f'{ping.db_name} ({ping.ping_time})'

    def pythonize_3_ping(self, ping_data):
        """ This method convert a ping coming from an Odoo instance running on python 2.
            Ping are a serialized dict which are then unserialized using ast.literal_eval on our side.
            The problem is that some values sent by the instances are computed using a 'search_count' which
            returns a long integer represented by a L at the end of the value.
            Python 3 doesn't support long integer anymore since all integer a longs.

            This regex check for the 4 fields that are computed using search_count and remove the L :D """
        pythonize_3_re = re.compile(r"('(?:nbr_users|nbr_active_users|nbr_share_users|nbr_active_share_users|IoTBox)':\s*\d+)L")
        return pythonize_3_re.sub(r'\g<1>', ping_data)

    @api.model
    def ping(self, arg0, ip_address):
        """ Create the database if this one does not exist.
            Store the ping from the server of the customer but with one limitation
            to avoid the ping flooding, one ping per day

            :param payload: a dictionnary from the controller in the front-end
                            (see controllers/main.py)
            :param ip: the IP address of the customer
        """
        def obfuscate_email(text):
            email = tools.email_normalize(text)
            if not email: return ""
            splited_mail = email.split("@")
            return "%s...@%s" % (splited_mail[0][:2], splited_mail[1])

        # @stw arg0 is always a string because it'a HTTP post or GET var
        arg0 = self.pythonize_3_ping(arg0)
        payload = ast.literal_eval(arg0)

        # Backward compatibility with 6.1 <field> -> company_<field>
        for old_field in 'name phone email'.split():
            value = payload.pop(old_field, None)
            if value:
                payload['company_%s' % old_field] = value

        Database = self.env["openerp.enterprise.database"]
        databases = Database.search([('db_uuid', '=', payload['dbuuid'])])

        # if the database does not exists then we create it
        if not databases:
            vals = {
                'db_uuid': payload['dbuuid'],
            }

            # Copy tags from the parent database
            parent_uuid = payload.get('parent_uuid', False)
            if parent_uuid:
                parent_db = Database.search([('db_uuid', '=', parent_uuid)])
                vals['tag_ids'] = [fields.Command.set(parent_db.tag_ids.ids)]

            databases = Database.create(vals)

        database = databases[0]

        # check if the last ping time is inferior of one day, to avoid a ping flooding
        if database.last_ping_id.ping_time:
            last_ping_time = database.last_ping_id.ping_time
        else:
            last_ping_time = datetime.datetime.min

        enterprise_code = payload.get('enterprise_code')

        if payload.get('force_ping'):
            can_store_ping = True
        elif enterprise_code and enterprise_code != database.subscription_id.client_order_ref:
            can_store_ping = True
        else:
            can_store_ping = last_ping_time + datetime.timedelta(days=1) < datetime.datetime.now()

        stored_ping_id = None
        if can_store_ping:
            _logger.debug("Regular ping for %r", payload)
        else:
            _logger.debug("DUPE ping for %r", payload)
            stored_ping_id = "DUPE"

        apps = payload.pop('apps', [])
        version = payload.get('version')
        database_values = {}
        if can_store_ping:
            ping_values = dict(
                database=database.id,
                ip=ip_address,
            )

            ping_values['db_name'] = payload.pop('dbname')
            payload_fields = """
                nbr_users version language nbr_active_users nbr_share_users nbr_active_share_users
                company_name company_email company_phone web_base_url website_homepage_modified
                last_cnx_date expire_date extra_apps
            """.split()
            for field in payload_fields:
                ping_values[field] = payload.pop(field, False)

            if not ping_values['language']:
                ping_values['language'] = 'en_US'

            # We update the app status (installed/uninstalled)
            database_app_ids = dict((app.module_id.id, app.id) for app in database.app_ids)

            # we search if the parsed version exists, if it doesn't we use the latest supported version
            version_id = self.env['openerp.enterprise.module.version'].search([('name', '=', humanize_version(version))]) or\
                         self.env['openerp.enterprise.module.version']._get_latest_version()

            app_ids = version_id.module_ids.filtered(lambda x: x.technical_name in apps).ids

            today_str = fields.Date.today()
            app_values = []
            for app_id in app_ids:
                if app_id in database_app_ids.keys():
                    app_values.append((1, database_app_ids[app_id], {'last_date': today_str, 'installed': True}))
                    database_app_ids.pop(app_id, False)
                else:
                    app_values.append((0, 0, {'module_id': app_id, 'installed': True, 'first_date': today_str, 'last_date': today_str}))
            for module_id, line_id in database_app_ids.items():
                app_values.append((1, line_id, {'installed': False}))

            database_values['app_ids'] = app_values

            database_users = dict((database_user.login, database_user) for database_user in database.database_user_ids)
            users = payload.pop('users', [])
            user_oauth_uids = payload.pop('odoo_oauth_uids', [])
            user_odoo_com_uids = []
            for user in users:
                # fix for when the user is putting an invalid value on the 'oauth_uid' field
                odoo_com_uid = user.pop('odoo_com_uid', False)
                try:
                    odoo_com_uid = int(odoo_com_uid)
                except ValueError:
                    odoo_com_uid = False

                if odoo_com_uid and self.env['res.users'].browse(odoo_com_uid).exists():
                    try:
                        user['user_id'] = odoo_com_uid
                        if user.pop('oauth', False):
                            user_oauth_uids.append(user['user_id'])
                        else:
                            user_odoo_com_uids.append(user['user_id'])
                    except ValueError:
                        pass

                if user.get('login'):
                    user_values = dict(
                        (field, user[field])
                        for field in ['name', 'login', 'password_crypt', 'login_date']
                        if user.get(field)
                    )
                    user_values['user_id'] = user.get('user_id', False)
                    if user['login'] in database_users:
                        database_user = database_users.pop(user['login'])
                        if database_user._is_different(user_values):
                            database_values.setdefault('database_user_ids', []).append((1, database_user.id, user_values))
                    else:
                        database_values.setdefault('database_user_ids', []).append((0, None, user_values))
            for remaining_user in database_users.values():
                database_values.setdefault('database_user_ids', []).append((2, remaining_user.id, None))

            # 12.0 adaptation of the oauth mechanism
            oauth_user_ids = set(user_oauth_uids + user_odoo_com_uids)
            to_update = database.registered_user_ids.filtered(lambda x: x.user_id.id in oauth_user_ids)
            to_remove = database.registered_user_ids - to_update
            database_values['registered_user_ids'] = []

            for u in to_remove:
                database_values['registered_user_ids'].append((2, u.id, 0))

            for u in users:
                if not u.get('user_id') or u['user_id'] not in oauth_user_ids:
                    continue
                # need to update existing record
                if u['user_id'] in to_update.mapped('user_id').ids:
                    oauth_values = {
                        'is_admin': u.get('is_admin', False),
                    }
                    oauth_user_id = to_update.filtered(lambda x: x.user_id.id == u['user_id'])[0]  # noqa: B023
                    # only update if needed
                    if oauth_user_id.is_admin != oauth_values['is_admin']:
                        database_values['registered_user_ids'].append((1, oauth_user_id.id, oauth_values))
                # create a new record
                else:
                    database_values['registered_user_ids'].append((0, 0, {
                        'user_id': u['user_id'],
                        'is_admin': u.get('is_admin', False),
                    }))

            info = dict(payload)
            for field in "contracts dbuuid id".split():
                info.pop(field, None)

            # some databases have an empty or space-only parent_uuid, and it should be ignored
            if payload.get('parent_uuid', '').strip():
                # however, the parent_uuid is copied from the database, so it should match exactly, including extra spaces
                database_values['parent_id'] = self.env['openerp.enterprise.database'].search([('db_uuid', '=', payload.get('parent_uuid'))]).id
            else:
                database_values['parent_id'] = False

            # if the database ping then we should set the visibility back, the database isn't anymore hidden
            database_values['visibility'] = 'support' if info.pop('db_purpose', None) == 'support' else 'customer'

            ping_values["info"] = json.dumps(info, indent=2)

            stored_ping_id = self.create(ping_values).id

            if database_values:
                database.write(database_values)

        contract = database.subscription_id
        contracts = {}
        enterprise_info = {}

        if enterprise_code:
            enterprise_code = enterprise_code.strip()
        if not contract or (enterprise_code and contract.client_order_ref != enterprise_code):
            # Used from 9.0 to link database with enterprise contract
            if enterprise_code:
                subscription = self.env['sale.order']._get_most_valid_subscription(enterprise_code)
                if subscription:
                    if subscription.enterprise_is_enterprise and subscription.database_ids:
                        # send back to the database information about why he cannot register his database with this contract code
                        # since the contract is already used in an other db, we warn him that he can unlink the other db from this contract
                        enterprise_info['database_already_linked_subscription_url'] = subscription._get_portal_url(access_token=False)
                        enterprise_info['database_already_linked_email'] = obfuscate_email(subscription.partner_id.commercial_partner_id.email)
                        enterprise_info['database_already_linked_send_mail_url'] = subscription._get_unlink_mail_send_url()
                        _logger.warning("Contract %s is already associated with database(s) %s", subscription.client_order_ref, subscription.database_ids.mapped('display_name'))
                    elif subscription.enterprise_is_enterprise or subscription.enterprise_is_partnership:
                        # A valid contract has been found for this db
                        contract = subscription
                        database.subscription_id = contract

        if can_store_ping:
            # update database validity based on contract status, etc.
            database.update_valid_date(code=enterprise_code)

        enterprise_info.update({
            'expiration_date': fields.Datetime.to_string(database.date_valid),
            'enterprise_code': contract and contract.client_order_ref or False,
        })

        # Contract data, used in 6.1 for registering contracts
        if contract:
            contract_dict = {
                'date_from': contract.start_date and fields.Datetime.to_string(contract.start_date),
                'date_to': contract.end_date and fields.Datetime.to_string(contract.end_date),
                # various states exists on 6.1 but only valid/unvalidated are meaningful
                'state': 'valid' if contract.subscription_state in ('3_progress', '4_paused') else 'unvalidated',
                'check_support': True,
                'check_opw': True,
                'kind': 'support',
            }
            contracts = {
                contract.client_order_ref: contract_dict
            }
            # Expiration date, used in 9.0 for registering enterprise
            if contract.enterprise_is_partnership:
                enterprise_info['expiration_reason'] = 'demo'
                enterprise_info['enterprise_code'] = False
            elif contract.enterprise_is_enterprise:
                reason = 'renewal'
                if can_store_ping and contract.subscription_state in ('3_progress', '4_paused'):
                    need_upsell = contract._plan2022_need_upsell()
                    reason = contract.set_upsell_status(need_upsell, reason)
                    # Need to check the upsell of the maintenance subscription in case there is one
                    mnt_contract = contract.maintenance_subscription_id
                    maintenance_need_upsell = mnt_contract._maintenance_need_upsell()
                    reason = mnt_contract.set_upsell_status(maintenance_need_upsell, reason)
                    database.update_valid_date()
                enterprise_info['expiration_date'] = fields.Datetime.to_string(database.date_valid)
                enterprise_info['expiration_reason'] = reason
        if not version.split('-')[0].endswith('e'):
            enterprise_info['expiration_date'] = enterprise_info['expiration_reason'] = False

        result = {
            'messages': [],  # retro-compatibility
            'contracts': contracts,
            'enterprise_info': enterprise_info,
        }
        _logger.info("Ping [%s] for database %s (%s) and code [%s], returning [%s], installed apps [%s]",
                     stored_ping_id, database.db_name, database.db_uuid, enterprise_code or 'n/a', result, apps)
        return repr(result)

    @api.model
    def _cron_pings_gc(self, limit=None):
        """
        Periodically delete old pings from the database:
        - between Y-1 and Y-5: keep only the pings of the first 8 days of each month
        - before Y-5: keep only the leftover pings in January
        (we prevent deletion of the first and last ping of a database)

        :param limit maximum number of pings we delete per cron invoke
        """
        limit_clause = 'LIMIT %s' if limit else ''
        args = [limit] if limit_clause else []
        self.env.cr.execute(f"""
            DELETE
            FROM openerp_enterprise_ping
            WHERE id IN (SELECT id
                         FROM openerp_enterprise_ping p
                         WHERE p.ping_time < CURRENT_DATE - INTERVAL '5 years'
                           AND EXTRACT(MONTH FROM p.ping_time) != 1 -- not in January
                           AND NOT EXISTS(SELECT NULL
                                          FROM openerp_enterprise_database db
                                          WHERE (db.first_ping = DATE(p.ping_time) AND db.id = p.database)
                                             OR db.last_ping_id = p.id)
                         UNION ALL
                         SELECT id
                         FROM openerp_enterprise_ping p
                         WHERE p.ping_time < CURRENT_DATE - INTERVAL '1 years'
                           AND EXTRACT(DAY FROM p.ping_time) > 8   -- not the first 8 days
                           AND NOT EXISTS(SELECT NULL
                                          FROM openerp_enterprise_database db
                                          WHERE (db.first_ping = DATE(p.ping_time) AND db.id = p.database)
                                             OR db.last_ping_id = p.id)
                         {limit_clause})  -- global limit
            """, args)
        _logger.info('ping garbage collect done, %s ping deleted', self.env.cr.rowcount)


class OpenerpEnterpriseTrialLog(models.Model):
    _name = "openerp.enterprise.trial.log"
    _description = "Information about a SaaS database installation"
    _order = "create_date desc"
    _rec_name = 'db_name'

    database_id = fields.Many2one('openerp.enterprise.database', readonly=True, auto_join=True)
    db_name = fields.Char(related="database_id.db_name", store=True, index="trigram")
    creation_date = fields.Datetime("Creation Date", readonly=True)
    hosting = fields.Char('Hosting', readonly=True)
    app_list = fields.Char('1st Apps Installed', readonly=True)
    lang_id = fields.Many2one('res.lang', string='Language', readonly=True)
    version = fields.Char('Version', readonly=True)
    country_id = fields.Many2one('res.country', string='Country', readonly=True)
    build_time = fields.Integer('Build time (s)', readonly=True, aggregator="avg")
    creation_method = fields.Char('Creation method', readonly=True)
    analytics_opt_in = fields.Boolean('Analytics opt-in', help='Whether the user agreed to UX analytics trackers being used.')

    last_ping = fields.Date(related='database_id.last_ping', store=True)
    mode = fields.Char(related='database_id.mode', store=True)
    extra_apps = fields.Boolean(related='database_id.extra_apps')
    has_contract = fields.Boolean(compute="_get_contract", search="_search_has_contract")

    ip_address = fields.Char('Origin IP address', index='btree_not_null')

    def _get_contract(self):
        for rec in self:
            rec.has_contract = bool(rec.database_id.subscription_id)

    def _search_has_contract(self, operator, value):
        if (operator == '=' and value) or (operator == '!=' and not value):
            return [('database_id.subscription_id', '!=', False)]
        else:
            return [('database_id.subscription_id', '=', False)]

    def _recaptcha_check_required(self):
        """ This method is used at the trial creation to
            decide if the Google reCAPTCHA score needs to be computed or not.

            To determine if the score needs to be computed, we retrieve the creation count
            of trial databases for the last 10 minutes.
            If this count exceeds a certain threshold set on the config parameter `trial_captcha_check_rate_10min`
            then we return True.

            Otherwise we do not check the reCAPTCHA scores. It is mainly used when we are attacked by bots.
        """
        check_rate = int(self.env['ir.config_parameter'].sudo().get_param('trial_captcha_check_rate_10min', '50'))

        creation_rate = self.sudo().search_count([
            ('create_date', '>=', fields.Datetime.now() - relativedelta(minutes=10))
        ])
        return creation_rate >= check_rate

    def _recaptcha_minimal_score(self):
        return float(self.env['ir.config_parameter'].sudo().get_param('trial_captcha_minimal_score', '0.5'))

    def _get_antispam_bypass_code(self, expiration=None):
        """ Generate a 12h temporary antispam bypass code, e.g. to provide a trial URL for a blocked customer,
            or when `expiration` is provided, produce the expected code for that expiration.
        """
        expiration = int(expiration or (time.time() + 12 * 3600))
        sig = odoo.tools.hmac(self.sudo().env, "odoo:trial:please:let:me:pass", expiration)
        return f"{sig}o{expiration}"

    def _verify_antispam_bypass_code(self, code):
        # Code is `hmac + o + expiry`, generate the code for given expiry and compare.
        expiry = code.split("o")[-1]
        still_valid = time.time() < int(expiry)
        return still_valid and odoo.tools.consteq(code, self._get_antispam_bypass_code(expiry))


class OpenerpEnterpriseDatabaseUser(models.Model):

    _name = "openerp.enterprise.database.user"
    _description = "OAuth user info between trial db <=> Accounts"

    def init(self):
        self.env.cr.execute("""
            DO $$
                BEGIN
                    ALTER TABLE openerp_enterprise_database_user ADD COLUMN password_crypt varchar;
                EXCEPTION
                    WHEN duplicate_column THEN null;
                END;
            $$;
        """)

    name = fields.Char('Name')
    login = fields.Char('Login', required=True)
    password_crypt = fields.Char('Encrypted password', compute='_compute_password', inverse='_inverse_password')
    user_id = fields.Many2one('res.users', 'Odoo.com User', ondelete='cascade')
    database_id = fields.Many2one('openerp.enterprise.database', 'Database', index=True, ondelete="cascade")
    to_sync = fields.Boolean('To synchronize', help='The credentials must be synchronized on the trial database')
    login_date = fields.Datetime(string='Latest connection')

    def _invite_users(self, saas_secret_signature, login, db_user_id, invitee):
        """ Method to invite users on a user's saas instance
            param str saas_secret_signature: A secret signature generated by the SaaS instance and valid only on this instance
            param int db_user_id: The user's id in the SaaS instance
            param list invitee: [{'name': XXX, 'mail': XXX}, ...]
        """
        self.ensure_one()
        db = self.database_id
        #
        params = {'params': {
                # the user login & id are needed for security check
                'login': login,
                'uid': db_user_id,
                # we need to provide a 'secret_signature' that can only be computed by the SaaS db
                'secret_signature': saas_secret_signature,
                'invitee': {i['mail']: i['name'] for i in invitee if len(i['mail']) > 0},
            }
        }

        try:
            # call the saas_trial route that create invitation on the SaaS db
            create_invitation_url = '%s://%s.%s/saas_trial/create_invitation' % (SCHEME, db.db_name, SAAS_BASE_URL)
            response = requests.post(create_invitation_url, json=params, auth=SaasAuth(), timeout=TIMEOUT)
            response.raise_for_status() # json rpc should only return 200
            result = response.json().get('result', {'result': False})
        except Exception:
            return False

        # if the users are successfuly invited
        if result.get('result') is True:
            # We need to commit the current transaction, because ping is called and alter openerp.enterprise.database.user
            # This avoid to raise a concurrent update issue
            self.env.cr.commit()
            already_invited_mails = self.gift_invited_mails and self.gift_invited_mails.split(',') or []
            self.gift_invited_mails = ','.join(set(already_invited_mails + [i['mail'] for i in invitee if len(i['mail']) > 0]))
            if self.gift_invitation_state == 'skipped':
                self.gift_invitation_state = 'pending'
            # Invitation email are sent by the SaaS database.
            return True

        return False

    def _compute_password(self):
        """ Prevent to read password """
        for db_user in self:
            db_user.password_crypt = '***'

    def _inverse_password(self):
        value = self[0].password_crypt
        self.env.cr.execute('UPDATE openerp_enterprise_database_user SET password_crypt = %s WHERE id in %s', (value, tuple(self.ids)))
        self.env.invalidate_all()


    def _is_different(self, values):
        result = dict.fromkeys(self.ids, True)
        self.env.cr.execute("""
            SELECT id, name, login, password_crypt, user_id
            FROM openerp_enterprise_database_user
            WHERE id in %s""", (tuple(self.ids),))
        for user in self.env.cr.dictfetchall():
            result[user['id']] = any(value != user[field] for field, value in values.items() if field in ['name', 'login', 'password_crypt', 'user_id'])
        return result


    def _verify(self, password):
        result = dict.fromkeys(self.ids, False)
        if self.ids:
            self.env.cr.execute("""
                SELECT id, password_crypt
                FROM openerp_enterprise_database_user
                WHERE id in %s AND password_crypt is not null""", (tuple(self.ids),))
            for user in self.env.cr.dictfetchall():
                result[user['id']] = enterprise_crypt_context.verify(password, user['password_crypt'])
        return result


class SaleReport(models.Model):
    _inherit = "sale.report"

    origin = fields.Char('Source Document', readonly=True)

    def _select_additional_fields(self):
        res = super()._select_additional_fields()
        res['origin'] = "s.origin"
        return res

    def _group_by_sale(self):
        res = super()._group_by_sale()
        res += """,
            s.origin"""
        return res


class ProductCategory(models.Model):
    _inherit = 'product.category'

    # xal. temp. while merged in standard
    l10n_in_hsn_code = fields.Char(string="HSN/SAC Code", help="Harmonized System Nomenclature/Services Accounting Code")


class ProductTemplate(models.Model):
    _inherit = 'product.template'

    l10n_in_hsn_code = fields.Char(string="HSN/SAC Code", help="Harmonized System Nomenclature/Services Accounting Code",
                                   compute="_compute_l10n_in_hsn_code", store=True)

    is_odoo_app = fields.Boolean(string='Is a paid Odoo App', compute='_compute_is_odoo_app')
    is_odoo_user = fields.Boolean(string='Is a paid Odoo User', compute='_compute_is_odoo_user')
    is_odoo_sh = fields.Boolean(string='Is a Odoo.sh product', compute='_compute_is_odoo_sh')

    @api.depends('categ_id.l10n_in_hsn_code')
    def _compute_l10n_in_hsn_code(self):
        for p in self:
            p.l10n_in_hsn_code = p.categ_id.l10n_in_hsn_code

    @api.depends('categ_id')
    def _compute_is_odoo_app(self):
        product_app_category = self.env.ref('openerp_enterprise.category_openerp_apps')
        product_odoo_sh_category = self.env.ref('openerp_enterprise.product_category_odoo_sh')
        extra_product = [
            self.env.ref('openerp_enterprise.product_multiyear_discount').product_tmpl_id,
            self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount').product_tmpl_id,
            self.env.ref('openerp_enterprise.product_learning_partner_discount').product_tmpl_id,
            self.env.ref('openerp_enterprise.product_official_partner_discount').product_tmpl_id,
        ]
        for p in self:
            p.is_odoo_app = p.categ_id in [product_app_category, product_odoo_sh_category] or p in extra_product


    @api.depends('categ_id')
    def _compute_is_odoo_user(self):
        product_user_category = self.env.ref('openerp_enterprise.product_category_odoo_users')

        product_ignored = [
            self.env.ref('openerp_enterprise.product_multiyear_discount').product_tmpl_id,  # keep uom / year
            self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount').product_tmpl_id,
            self.env.ref('openerp_enterprise.product_learning_partner_discount').product_tmpl_id,
            self.env.ref('openerp_enterprise.product_official_partner_discount').product_tmpl_id,
            self.env.ref('openerp_enterprise.product_prorated_discount').product_tmpl_id,  # keep PCE
        ]

        for p in self:
            p.is_odoo_user = (p.categ_id == product_user_category) and p not in product_ignored

    @api.multi
    def _compute_is_odoo_sh(self):
        product_odoo_sh_category = self.env.ref('openerp_enterprise.product_category_odoo_sh')
        for p in self:
            p.is_odoo_sh = p.categ_id == product_odoo_sh_category


class ProductProduct(models.Model):
    _inherit = 'product.product'

    def _is_add_to_cart_allowed(self):
        self.ensure_one()

        allowed_categories_xmlids = [
            'openerp_enterprise.product_category_odoo_users', # Odoo Users
            'openerp_enterprise.category_openerp_apps', # Odoo Apps
            'openerp_enterprise.product_category_partnership', # Partnership
            'openerp_enterprise.category_custom_apps', # Online Custom
            'openerp_enterprise.product_category_odoo_sh', # Odoo.sh products
            'openerp_enterprise.category_maintenance', # Maintenance product
        ]

        return super()._is_add_to_cart_allowed() or self.sudo().categ_id in [self.env.ref(xmlid) for xmlid in allowed_categories_xmlids]


class DatabaseRestoreWizard(models.TransientModel):
    _name = 'database.restore.wizard'
    _description = 'Database Restore Wizard'

    src_dbname = fields.Char('Source Database Name', required=True, readonly=True)
    dst_dbname = fields.Char('New Database Name', required=True)
    db_uuid = fields.Char('Database UUID', required=True)
    hosted_on_saas = fields.Boolean('Hosted On SaaS', required=True, readonly=True)
    operation = fields.Selection([
        ('copy', '[TESTING] Duplicate'),
        ('recovery', '[PRODUCTION] Recover database'),
        ('disaster', '[DISASTER] Recover database and its metadata'),
    ], default='copy', required=True)
    snapshot = fields.Char('Snapshot', help='If not set, the last snapshot is selected')
    user_id = fields.Many2one(
        'res.users', string='Send Notification To',
        domain=lambda self: [('groups_id', 'in', self.env.ref('openerp_enterprise.group_support_functional').id)],
        default=lambda self: self.env.user.id
    )
    snapshot_list = fields.Text("Available snapshots", readonly=True)

    def _dst_dbname_exists(self):
        """
        Check if the dst_dbname already exists on odoo.com. If we get a redirect (= status code
        302), the DB is considered not existing. In other cases, it is considered existing.
        """

        url = "https://{}.{}/saas_worker/noop".format(self.dst_dbname, SAAS_BASE_URL)
        res = requests.get(url, allow_redirects=False, auth=SaasAuth(), timeout=TIMEOUT)
        if res.status_code == 302:
            return False
        return True

    def action_request_restore(self):
        self.ensure_one()
        if not self.hosted_on_saas:
            raise UserError('Restricted to databases hosted on SaaS')
        if not self.env.user.has_group('openerp_enterprise.group_support_functional'):
            raise AccessDenied('Restricted to functional/technical support staff')

        # In disaster mode, we import the metadata, therefore we must keep the same name.
        if self.operation == 'disaster' and self.dst_dbname != self.src_dbname:
            raise UserError(_(
                "In disaster mode, the new database name must be identical to the source database name. "
                "If a new name is expected, the DB should be renamed after restore."
            ))

        # Prevent from restoring a DB in production mode if it is reachable.
        if self.operation != 'copy' and self._dst_dbname_exists():
            raise UserError(_(
                'Database %s exists and is reachable! To use this mode, the database must not exist.\n'
                'Suggestion: duplicate the production database, drop it then try again.',
                self.dst_dbname,
            ))

        payload = {
            'original_name': self.src_dbname,
            'db_uuid': self.db_uuid,
            'snapshot': self.snapshot or '',
            'target_db_name': self.dst_dbname,
            'accounts_user_id': self.user_id.id or self.env.user.id,
            'operation': self.operation,
        }

        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', self.db_uuid)])
        reason = f"db: {database.db_name}"
        if not database._check_and_log_spoof('restore', reason=reason):
            raise AccessDenied(_("This database is restricted! You need to ask for a staff access authorization to the customer."))

        res = requests.post(
            f"{SAAS_OPS_URL}/backup_db/saas/restore",
            data=json.dumps(payload),
            headers={"Content-type": "application/json"},
            auth=SaasAuth(),
        )
        response = res.json()
        _logger.info('action_request_restore: database restore requested by %s (id=%s): payload=%s, response=%s', self.env.user.login, self.env.user.id, payload, response)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code == 400:
                raise UserError("Error during restore request:\n{}".format(
                    response.get("error"))
                ) from e
            else:
                raise UserError(
                    "Error during restore request ({})".format(e.response.status_code)
                ) from e


    @api.model
    def default_get(self, default_fields):
        res = super(DatabaseRestoreWizard, self).default_get(default_fields)
        databases = self.env['openerp.enterprise.database'].browse(self.env.context['active_ids'])
        if len(databases) == 1:
            res['src_dbname'] = res['dst_dbname'] = databases.db_name
            res['db_uuid'] = databases.db_uuid
            res['hosted_on_saas'] = databases.hosted_on_saas
            if res['db_uuid'] and res['hosted_on_saas']:
                res['snapshot_list'] = self._get_snapshot_list(res['db_uuid'])
        return res

    def _get_snapshot_list(self, db_uuid):
        try:
            response = requests.get(
                SAAS_OPS_URL + "/backup_db/saas/backup/{}".format(db_uuid),
                auth=SaasAuth(),
                timeout=10,
            )
            response.raise_for_status()
            backups = response.json()
            return ", ".join([b["snapshot"] for b in backups])
        except Exception as e:
            _logger.warning("Error while fetching backups: %s", str(e))
            return "Unable to fetch snapshot list"
